import java.io.File

fun main() {
    val input = File("input")
    val histories = parseHistories(input)
    println(part1(histories))
    println(part2(histories))
}

fun part1(histories: List<List<Int>>): Int {
    return histories.map { extrapolate(it) }.sum()
}

fun part2(histories: List<List<Int>>): Int {
    return histories.map { extrapolate(it.reversed()) }.sum()
}

fun extrapolate(history: List<Int>): Int {
    tailrec fun recurse(history: List<Int>, last: Int): Int {
        if (history.all { it == 0 }) {
            return 0 + last
        }

        val future = history.windowed(2) { (a, b) -> b - a }
        return recurse(future, last + future.last())
    }

    return recurse(history, history.last())
}

fun parseHistories(input: File): List<List<Int>> {
    return input.useLines { it.map { it.split(' ').map { it.toInt() } }.toList() }
}
